class ColumnType:
    """
    column type class, used to identify column types 
    """
    def __init__(self, col_type):
        self.col_type = col_type

        self._syntax = col_type 

    def __str__(self):
        return self._syntax

class UnaryColumnType(ColumnType):
    """
    column type that takes single operand like varchar(10).
    instantiated with column type as input.
    instance is callable with operand as input.
    """
    def __init__(self, col_type):
        super().__init__(col_type)

    def __call__(self, operand):
        self._syntax = f'{self.col_type}({operand})'
        return self

class BinaryColumnType(ColumnType):
    """
    column type that takes two operands like decimal(10, 25)
    instantiated with column type as input
    instance is ballable with operands as inputs 
    """
    def __init__(self, col_type):
        super().__init__(col_type)

    def __call__(self, operand1, operand2):
        self._syntax = f'{self.col_type}({operand1}, {operand2})'
        return self

class _COLUMNTYPE:
    boolean = ColumnType('boolean')
    tinyint = ColumnType('tinyint')
    smallint = ColumnType('smallint')
    int = ColumnType('int')
    bigint = ColumnType('bigint')
    double = ColumnType('double')
    float = ColumnType('float')
    string = ColumnType('string')
    binary = ColumnType('binary')
    date = ColumnType('date')
    timestamp = ColumnType('timestamp')
    char = UnaryColumnType('char')
    varchar = UnaryColumnType('varchar')
    decimal = BinaryColumnType('decimal')
    #TODO array, map, struct

COLUMNTYPE = _COLUMNTYPE()



import copy

from athenaSQL.queries.select import SelectQuery
from athenaSQL.queries.query_abc import QueryABC, \
    _exit_on_uncalled_preceding_method


class CTEQuery(QueryABC):
    """
    """

    def __init__(self):
        self._temp_tables = {}
        self._select_query = None

    def _to_sql(self):
        pass

    def withTable(self, table_name, select_query):
        """
        """

        # TODO table_name validation

        if not isinstance(select_query, SelectQuery):
            raise TypeError(f'{type(select_query).__name__} is not a type '
                            f'of SelectQuery')

        clone_obj = copy.deepcopy(self)
        self._temp_tables[table_name] = select_query
        clone_obj._temp_tables = self._temp_tables

        return clone_obj

    def table(self, table_name):
        """
        """
        # raise error there is not table to select from
        _exit_on_uncalled_preceding_method(self._temp_tables,
                                           'table',
                                           'withTable')
        # TODO table_name validation

        # raise error if selected table is not in temp tables
        if table_name not in self._temp_tables.keys():
            raise ValueError(f'temporary table `{table_name}` does not exist')

        _query = SelectQuery(table_name)
        _query._cte_tables = self._temp_tables

        return _query


def withTable(table_name, select_query):
    """
    constract a CTE query
    """
    return CTEQuery().withTable(table_name, select_query)



import copy
from jinja2 import Template

from athenaSQL.queries.query_abc import QueryABC, \
    _exit_on_override, \
    _exit_on_partial_query, \
    _check_and_extract_list_or_valid_typed_arguments

from athenaSQL.orders import ASC, DESC
from athenaSQL.column import Column

window_query_template = """
OVER (
    {%- if partitions -%}
    PARTITION BY
    {%- for partition in partitions -%}
    {{' '}}{{ partition }}
    {%- if not loop.last -%}
    ,
    {%- endif -%}
    {%- endfor -%}
    {%- endif -%}
    {%- if orders %}
      ORDER BY 
    {%- for order in orders -%}
    {{' '}}{{ order }}
    {%- if not loop.last -%}
    ,
    {%- endif -%}
    {%- endfor -%}
    {% endif %})
"""


class WindowQuery(QueryABC):
    """
    A query class to constract to constract the clause after and including
    over syntax in a sql window function.

    >>> win_func() OVER (PARTITION BY col1 ORDER BY col2)

    this class constract `OVER (PARTITION BY col1 ORDER BY col2)` caluse
    """

    def __init__(self):
        self._partitions = []
        self._orders = []

    def _to_sql(self):

        # check if query constraction is complete
        _exit_on_partial_query(
            (self._partitions or self._orders),
            'partitionBy', 'WindowQuery')

        return Template(window_query_template).render(
            partitions=self._partitions,
            orders=self._orders
        ).strip()

    def partitionBy(self, *cols):
        """
        """
        # check if method is already been called in an instance
        _exit_on_override(self._partitions, 'partitionBy',
                          'partitionBy(*cols)')

        arguments = _check_and_extract_list_or_valid_typed_arguments(
            cols, 'partitionBy')

        clone_obj = copy.deepcopy(self)
        clone_obj._partitions = arguments

        return clone_obj

    def orderBy(self, *cols):
        """
        """
        # check if method is already been called in an instance
        _exit_on_override(self._orders, 'orderBy', 'orderBy(*cols)')

        arguments = _check_and_extract_list_or_valid_typed_arguments(
            cols, 'orderBy', valid_types=(str, ASC, DESC))

        # should also be done in Column types
        ordering_cols = list(map(lambda arg: arg
                                 if isinstance(arg, (ASC, DESC))
                                 else Column(arg),
                                 arguments))

        clone_obj = copy.deepcopy(self)
        clone_obj._orders = ordering_cols

        return clone_obj


Window = WindowQuery()



from athenaSQL.column import validate_column_name


class ASC:
    def __init__(self, column):
        self._column = validate_column_name(column)

    def __str__(self):
        return f'{self._column} DESC'


class DESC:
    def __init__(self, column):
        self._column = validate_column_name(column)

    def __str__(self):
        return f'{self._column} ASC'



from abc import ABC, abstractmethod

from athenaSQL.errors import PartialQueryError
from athenaSQL.column import Column, AliasColumn


def _exit_on_uncalled_preceding_method(method_called,
                                       method_name,
                                       preceding_method_name):
    """
    throw exception if preceding method is not called.
    params
    ------
    method_called: any
        a variable if not satisfy negation indicates preceding method has been
        called

    method_name: string
        function name of calling method 

    preceding_method_name: string
        function name of preceding method 
    """

    if not method_called:
        raise SyntaxError(f".{method_name}() cannot be called before "
                          f".{preceding_method_name}()")


def _exit_on_partial_query(query_attribute,
                           method_name,
                           query_name):
    """
    throw PartialQueryError if query_attribute satisfy nagation
    """

    if not query_attribute:
        raise PartialQueryError(
            f"{method_name}() is not called on {query_name}")


def _exit_on_override(previous_value, sql_clause, method_name):
    # exit if the previous value is not empty
    # this prevent overriding & multiple call of certain methods on chain
    if previous_value:
        raise AttributeError(f"You can't reassign {sql_clause} in a query, "
                             f"Use .{method_name} only once in your query")


def _check_and_extract_list_or_valid_typed_arguments(args,
                                                     method_name, valid_types=(str, Column, AliasColumn)):

    if not isinstance(valid_types, tuple):
        valid_types = (valid_types,)

    try:
        # argument is a single type, check for valid type and return
        if isinstance(args, valid_types):
            return args

        # argument is a list
        # check and return if all parameters are with in valid types
        if all(isinstance(arg, valid_types) for arg in args):
            return list(args)

        # check if there is a single parameter and of a type list
        if len(args) == 1 and isinstance(args[0], list):
            # check all items in the list valid types
            if all(isinstance(arg, valid_types) for arg in args[0]):
                return args[0]  # return the first parameter of a type list
    except:
        # escape if there are any errors, error message is raised below
        pass

    # if condition above doesn't fulfilled raise an error
    raise TypeError(f".{method_name}() can only accept a list or "
                    f"[{', '.join(map(lambda t: t.__name__, valid_types))}]")


class QueryABC(ABC):
    """
    abstract class for all query types.
    all query class should accept database name, table names and _to_sql method
    to render sql
    """

    def __init__(self, database, table):
        self.database = database
        self.table = table
        self.query_engine = None

        ABC.__init__(self)

    @abstractmethod
    def _to_sql(self, **kwargs):
        pass

    def show_query(self):
        """preivew sql query
        """
        print(self._to_sql())

    # TODO pkg athena and add as dependency
    # def exec(self, return_type='dataframe', dry_run=False):
    #     """
    #     run athena query, if dry_run is true return query string
    #     """
    #     query = self._to_sql()

    #     if dry_run:
    #         return self._to_sql()

    #     self.query_engine =  RunAthenaQuery(query=query, return_type=return_type)

    #     return self.query_engine.query_results()



from athenaSQL.queries import SelectQuery, InsertQuery, \
                                CreateQuery, CreateAsQuery

class AthenaTable:
    """
    abstract class for athena table
    """

    def __init__(self, *, database, table):
        self.database = database
        self.table = table

    def select(self, *cols):
        return SelectQuery(self.table, database=self.database) \
                .select(*cols)

    def insert(self, select_query):
        return InsertQuery(self.database, self.table) \
                .insert(select_query)

    def create(self):
        return CreateQuery(self.database, self.table)

    def createAs(self, select_query):
        return CreateAsQuery(self.database, self.table) \
                .createAs(select_query)

class TempTable:
    """
    abstract class for temporary table. 
    used to represent cte tables
    """

    def __init__(self, table):
        self.table = table

    def select(self, *cols):
        return SelectQuery(self.table).select(*cols)



def _exit_on_invalid_operand(other):
    """Exit if operand is not of a type int, str or Column
    """
    from athenaSQL.column import Column

    if not isinstance(other, (int, str, Column)):
        raise TypeError(
            f"Column can't be compared with {type(other).__name__}")


def col_boolOps(boolOps, compound=False):
    """
    A decorator to check unsupported operands for [&,|] ops and 
    change return type to ConditionalColumn if ops is bool 
    """
    def decorator(func):
        def wrapper(self, other):
            obj = func(self, other)

            from athenaSQL.column import ConditionalColumn
            # if operand is not ConditionalColumn and operation is boolOps
            if boolOps and compound and type(self) != ConditionalColumn:
                raise TypeError(f"unsupported operand type(s): "
                                f"'{type(self).__name__}' and "
                                f"'{type(other).__name__}'")

            if boolOps:
                return ConditionalColumn(self)

            return obj

        return wrapper

    return decorator


def _bin_op(operator, r=False, compound=False, boolOps=False):
    """Create a method for given binary operation
    """

    @col_boolOps(boolOps, compound=compound)
    def _(self, other):
        _exit_on_invalid_operand(other)

        # if operand is string enclose it in single quotes
        if isinstance(other, str):
            other = f"'{other}'"

        # self is on the left, operand is on the right
        # sum('col') + 6
        clause = f'{self._sql_clause} {operator} {other}'

        # self is on the right, operand is on the left
        # 6 + sum('col')
        if r:
            clause = f'{other} {operator} {self._sql_clause}'

        # operation is compound, it needs to be enclosed with parenthesis
        # (col > 5 AND col2 = 10)
        if compound:
            clause = f'({clause})'

        self._sql_clause = clause

        return self
    return _


def _func_op(operator, boolOps=False):
    """
    Create a method for given function operations

    Example
    -------
    >>> from adflow.sql.function import avg
    >>> ~(avg('col') > 5)
    NOT(AVG(col) > 5)
    """

    def _(self):
        from athenaSQL.column import ConditionalColumn
        # if operand is not ConditionalColumn and operation is boolOps
        if boolOps and type(self) != ConditionalColumn:
            raise TypeError(f'unsupported operand type: {type(self).__name__}')

        self._sql_clause = f'{operator}({self._sql_clause})'

        # if operation is conditional or logical return ConditionalColumn type
        if boolOps:
            return ConditionalColumn(self)

        return self
    return _


def _bin_func_op(operator, r=False, boolOps=False):
    """
    """

    @col_boolOps(boolOps)
    def _(self, other):
        _exit_on_invalid_operand(other)

        if isinstance(other, str):
            other = f"'{other}'"

        # r is False, self is the left operand of the binary funciton operator
        # POWER(SUM(col), 2)
        clause = f'{operator}({self._sql_clause}, {other})'

        # self is the right operand of the binary function operation
        # POWER(2, SUM(col))
        if r:
            clause = f'{operator}({other}, {self._sql_clause})'

        self._sql_clause = clause

        return self
    return _


class ComparisonMixin(object):
    """
    A mixin class to enable comparison operators on SQLFunctions

    Example
    -------
    >>> From adflow.interface.function import sum
    >>> str(sum('col'))
    SUM('col')
    >>> str(sum('col') >= 10)
    SUM('col') >= 10
    """

    # comparison operators
    __lt__ = _bin_op('<', boolOps=True)
    __le__ = _bin_op('<=', boolOps=True)
    __eq__ = _bin_op('=', boolOps=True)
    __gt__ = _bin_op('>', boolOps=True)
    __ge__ = _bin_op('>=', boolOps=True)
    __ne__ = _bin_op('<>', boolOps=True)


class ArithmeticMixin(object):
    """
    A mixin class to enable arithmetic operations on SQLFunctions

    Example
    -------
    >>> From adflow.interface.function import sum
    >>> str(sum('col'))
    SUM('col')
    >>> str(sum('col') + 10)
    SUM('col') + 10
    """

    # arithmetic operations
    __add__ = _bin_op("+")
    __sub__ = _bin_op("-")
    __mul__ = _bin_op("*")
    __mod__ = _bin_op("%")
    __truediv__ = _bin_op("/")
    __radd__ = _bin_op("+", r=True)
    __rsub__ = _bin_op("-", r=True)
    __rmul__ = _bin_op("*", r=True)
    __rmod__ = _bin_op("%", r=True)
    __rtruediv__ = _bin_op("/", r=True)

    # functional arithmetic operations
    __pow__ = _bin_func_op("POWER")
    __rpow__ = _bin_func_op("POWER", r=True)
    __abs__ = _func_op("ABS")

    # TODO __round__
    # TODO __trunc__
    # TODO __floor__
    # TODO __ceil__


class LogicalMixin(object):
    """
    A mixin class to enable logical arithmetic operators on SQLFunctions

    Example
    -------
    >>> FROM adflow.sql.function import sum, avg
    >>> str(sum('col'))
    SUM('col')
    >>> str(sum('col') == 10 AND avg('col2') == 5)
    (SUM('col') = 10 AND AVG('col2') = 5)
    """

    # logical operators
    __and__ = _bin_op("AND", compound=True, boolOps=True)
    __or__ = _bin_op("OR", compound=True, boolOps=True)
    __invert__ = _func_op("NOT", boolOps=True)
    __rand__ = _bin_op("AND", r=True, compound=True, boolOps=True)
    __ror__ = _bin_op("OR", r=True, compound=True, boolOps=True)



from athenaSQL.column import Column, NewColumn, \
    ConditionalColumn, CaseColumn, WindowColumn

from athenaSQL.queries.query_abc import (
    _check_and_extract_list_or_valid_typed_arguments)


def _wrap_col(col, wrapper, parent_wrapper):
    """
    wrap `col` with a `wrapper` class. `wrapper` should be of a type `wrapper_type`.
    """
    if not issubclass(wrapper, parent_wrapper):
        raise TypeError(f'{wrapper} is not a type of {parent_wrapper}')

    return wrapper(col)


def _create_unary_function(name, doc="", func_type=None):
    """ Create a function for unary function by name"""
    def _(col):

        # if col is string constract column before operation
        if isinstance(col, str):
            col = Column(col)

        if not isinstance(col, Column):
            raise TypeError(f'{col} argument is not type of a `Column`')

        col._sql_clause = (f'{name.upper()}({col})')

        # wrap col with custom functional column type
        if func_type:
            col = func_type(col)
            #col = _wrap_col(col, func_type, FunctionalColumn)

        return col

    _.__name__ = name
    _.__doc__ = doc
    return _


def _create_binary_function(name, doc="", func_type=None):
    """Create a function for binary function by name"""
    def _(col, operand):

        # if col is string constract column before operation
        if isinstance(col, str):
            col = Column(col)

        if not isinstance(col, Column):
            raise TypeError(f'{col} argument is not type of a `Column`')

        # enclose string arguments with single quotes
        # FUNC(col, operand), FUNC('col', 'operand')
        col._sql_clause = (f'{name.upper()}({col}, '
                           + (f'\'{operand}\'' if isinstance(operand,
                              str) else str(operand))
                           + ')')

        # wrap col with custom functional column type
        if func_type:
            col = func_type(col)
            #col = _wrap_col(col, func_type, FunctionalColumn)

        return col

    _.__name__ = name
    _.__doc__ = doc
    return _


def _create_nullnary_function(name, doc=""):
    """ """
    # TODO nullary functions like pi(), e()
    pass


def _create_window_function(name, doc=""):
    """Create  """
    def _():
        wCol = WindowColumn(name)
        return wCol

    _.__name__ = name
    _.__doc__ = doc
    return _


_unary_functions = {
    'sqrt': 'Computes the square root of the specified float value.',
    'abs': 'Computes the absolute value.',
    'mean': 'Aggregate function: returns the average of the values in a group.',
    'geometric_mean': 'Returns the geometric mean of all input values.',
    'stddev': 'Returns the sample standard deviation of all input values.',
    'variance': 'Returns the sample variance of all input values.',
    # Unicode Functions
    # opt argument
    'normalize': 'Transforms string with NFC normalization form.',
    'to_utf8': 'Encodes string into a UTF-8 varbinary representation.',
    # opt argument
    'from_utf8': 'Decodes a UTF-8 encoded string from binary. Invalid UTF-8 '
                 'sequences are replaced with the Unicode replacement character U+FFFD.',
    # Date and Time Functions
    'date': 'This is an alias for CAST(x AS date).',
    'last_day_of_month': 'Returns the last day of the month.',
    'from_iso8601_timestamp': 'Parses the ISO 8601 formatted string into a '
                              'timestamp with time zone.',
    # Date and Time Convenience Extraction Functions
    'day': 'Returns the day of the month from x.',
    'day_of_month': 'This is an alias for day().',
    'day_of_week': 'Returns the ISO day of the week from x. The value ranges '
                   'from 1 (Monday) to 7 (Sunday).',
    'day_of_year': 'Returns the day of the year from x. The value ranges from 1 to 366.',
    'dow': 'This is an alias for day_of_week().',
    'doy': 'This is an alias for day_of_year().',
    'hour': 'Returns the hour of the day from x. The value ranges from 0 to 23.',
    'millisecond': 'Returns the millisecond of the second from x.',
    'minute': 'Returns the minute of the hour from x.',
    'month': 'Returns the month of the year from x.',
    'quarter': 'Returns the quarter of the year from x. The value ranges from '
               '1 to 4.',
    'second': 'Returns the second of the minute from x.',
    'timezone_hour': 'Returns the hour of the time zone offset from timestamp.',
    'timezone_minute': 'Returns the minute of the time zone offset from timestamp.',
    'week': 'Returns the ISO week of the year from x. The value ranges from 1 to 53.',
    'week_of_year': 'This is an alias for week().',
    'year': 'Returns the year from x.',
    'year_of_week': 'Returns the year of the ISO week from x.',
    'yow': 'This is an alias for year_of_week().',
    # URL Functions
    # [protocol:][//host[:port]][path][?query][#fragment]
    'url_extract_fragment': 'Returns the fragment identifier from url.',
    'url_extract_host': 'Returns the host from url.',
    'url_extract_path': 'Returns the path from url.',
    'url_extract_port': 'Returns the port number from url.',
    'url_extract_protocol': 'Returns the protocol from url.',
    'url_extract_query': 'Returns the query string from url.',
    'url_encode': '',
    'url_decode': 'Unescapes the URL encoded value. This function is the '
                  'inverse of url_encode().',
    # UUID Functions
    'uuid()': 'Returns a pseudo randomly generated UUID (type 4).',

}

_binary_functions = {
    'nullif': 'Returns null if value1 equals value2, otherwise returns value1.',
    'mod': 'Returns the modulus (remainder) of n divided by m.',
    'power': 'Returns x raised to the power of p.',
    'pow': 'This is an alias for power().'
}

for _name, _doc in _unary_functions.items():
    globals()[_name] = _create_unary_function(_name, _doc)

for _name, _doc in _binary_functions.items():
    globals()[_name] = _create_binary_function(_name, _doc)

# Aliases for creating a column class by name
col = column = lambda col_name: Column(col_name)
col.__doc__ = column.__doc__ = "Create a column by name."

# Aliase for creating a new column with type


def nCol(col_name, data_type): return NewColumn(col_name, data_type)


def when(condition, value):
    """
    Start a case condition. Returns a CaseColumn 
    """
    if not isinstance(condition, ConditionalColumn):
        raise TypeError('condition argument in when() should be a '
                        'ConditionalColumn')

    return CaseColumn((condition, value))


def cast(col, _type):
    """Alias for Column().cast()."""

    # if col is string constract column before operation
    if isinstance(col, str):
        col = Column(col)

    if not isinstance(col, Column):
        raise TypeError('col argument in cast() should be a Column')

    return col.cast(_type)


def try_cast(col, _type):
    """Like cast(), but returns null if the cast fails."""

    # if col is string constract column before operation
    if isinstance(col, str):
        col = Column(col)

    if not isinstance(col, Column):
        raise TypeError('col argument in try_cast() should be a Column')

    return col.cast(_type, _try=True)



from athenaSQL.column import Column, AggregateColumn

from athenaSQL.functions.functions import _create_unary_function, \
    _create_binary_function

_unary_functions = {}

_binary_functions = {}

for _name, _doc in _unary_functions.items():
    globals()[_name] = _create_unary_function(_name, _doc)

for _name, _doc in _binary_functions.items():
    globals()[_name] = _create_binary_function(_name, _doc)


def round(col, d=None):
    """
    Returns col rounded to d decimal places.
    Round col to the nearest integer if d is not specified.
    """

    if not isinstance(col, (Column, str)):
        raise TypeError('col argument in replace should be a type of '
                        '[Column|str]')

    if isinstance(col, str):
        col = Column(col)

    col._sql_clause = f"ROUND({col}"
    col._sql_clause += f", {d})" if d else ")"

    return col



from athenaSQL.column import AggregateColumn

from athenaSQL.functions.functions import _create_unary_function, \
    _create_binary_function, \
    _create_nullnary_function

_unary_functions = {
    'any_value': 'Returns an arbitrary non-null value x, if one exists',
    'arbitrary': 'Returns an arbitrary non-null value of x, if one exists. Identical to any_value().',
    'array_agg': 'Returns an array created from the input x elements.',
    'avg': 'Returns the average (arithmetic mean) of all input values.',
    'bool_and': 'Returns TRUE if every input value is TRUE, otherwise FALSE.',
    'bool_or': 'Returns TRUE if any input value is TRUE, otherwise FALSE.',
    'checksum': 'Returns an order-insensitive checksum of the given values.',
    'count': 'Returns the number of non-null input values.',
    'count_if': 'Returns the number of TRUE input values. This function is equivalent to count(CASE WHEN x THEN 1 END).',
    'every': 'This is an alias for bool_and().',
    'geometric_mean': 'Returns the geometric mean of all input values.',
    'max': 'Returns the maximum value of all input values.',
    'min': 'Returns the minimum value of all input values.',
    'sum': 'Returns the sum of all input values.'
}

_binary_functions = {}

for _name, _doc in _unary_functions.items():
    globals()[_name] = _create_unary_function(_name, _doc, AggregateColumn)

for _name, _doc in _binary_functions.items():
    globals()[_name] = _create_binary_function(_name, _doc, AggregateColumn)



from importlib.metadata import version

from .athena import Athena
from .athena_table import AthenaTable, TempTable
from .column import Column
from .errors import PartialQueryError

__version__ = version(__name__)
